import 'dart:io';
import 'dart:async';
import 'package:adeptutils/adeptutils.dart' as AdeptUtils;
import 'package:chopper/chopper.dart';
import 'package:adeptarch_base/src/api/adept_chopper_client.dart';
import 'package:http/http.dart' as http;

final restAPI = RestAPI();

class RestAPI {
  AdeptChopperClient chopperClient;
  late Uri apiUri;
  http.Client? httpClient;
  Iterable interceptors;
  Authenticator? authenticator;
  Converter? converter;
  ErrorConverter? errorConverter;
  Map<Type, ChopperService> services;
  Duration? timeout;

  RestAPI({
    AdeptChopperClient? chopperClient,
    String? apiUrl,
    Uri? apiUri,
    this.httpClient,
    Iterable? interceptors,
    this.authenticator,
    Converter? converter,
    this.errorConverter,
    Map<Type, ChopperService>? services,
    this.timeout,
  })  : chopperClient = chopperClient ?? AdeptChopperClient(),
        apiUri = apiUri ?? Uri.parse(apiUrl ?? ''),
        interceptors = interceptors ?? [],
        converter = converter ?? const JsonConverter(),
        services = services ?? {};

  set apiUrl(String apiUrl) => apiUri = Uri.parse(apiUrl);
  String get apiUrl => apiUri.toString();

  void init() {
    try {
      chopperClient.dispose();
    } catch (e) {
      print(e);
    }
    chopperClient = AdeptChopperClient(
      baseUrl: apiUri,
      client: httpClient,
      interceptors: interceptors,
      authenticator: authenticator,
      converter: converter,
      errorConverter: errorConverter,
      services: services.values,
      timeout: timeout,
    );
    initServices();
  }

  void initServices() {
    services.forEach((type, service) {
      service.client = chopperClient;
    });
  }

  void dispose() {
    try {
      // _services.forEach((t, s) => s.dispose()); // ChopperService no longer support .dispose()
      services.clear();
      chopperClient.dispose();
    } catch (e) {
      print(e);
    }
  }

  void addService(ChopperService service) {
    service.client = chopperClient;
    services[service.definitionType] = service;
  }

  T service<T extends ChopperService>(Type type) {
    final s = services[type];
    if (s == null) {
      throw Exception("Service of type '$type' not found.");
    }
    return s as T;
  }

  ///Implement this method when checked what exceptions could be
  String? getLocalizedMessageFromException(dynamic exception) {
    if (exception == null) return null;
    if (exception is AdeptUtils.GenericException) {
      return getLocalizedMessageFromException(exception.causeException);
    }

    if (exception is WebSocketException ||
        exception is TimeoutException ||
        exception.toString().toLowerCase().contains("TimeoutException")) {
      return "Server connection timeout.";
    }
    if (exception is SocketException ||
        exception.toString().toLowerCase().contains("failed to connect")) {
      return "Check your internet connection. ${exception.toString()}";
    }
    if (exception is HandshakeException) return exception.toString();

    return exception.toString();
  }

  String? getGenericLocalizedMessageFromException(dynamic exception) {
    String? message = getLocalizedMessageFromException(exception);
    return message ?? "Unknown error";
  }

  String? getLocalizedMessageFromHttpResultCode(int? code) {
    if (code == null) return null;
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content";

      case HttpStatus.badRequest:
        return "($code) Bad Request";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized";

      case HttpStatus.forbidden:
        return "($code) Forbidden";

      case HttpStatus.notFound:
        return "($code) Not Found";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error";

      default:
        return "HTTP error code: ($code)";
    }
  }

  String? getLocalizedMessageFromResponse({Response? response, int? code}) {
    code ??= response?.statusCode;
    if (code == null) return null;
    String extraMessage = " ${response?.error ?? ""}";
    switch (code) {
      case HttpStatus.noContent:
        return "($code) No Content.$extraMessage";

      case HttpStatus.badRequest:
        return "($code) Bad Request.$extraMessage";

      case HttpStatus.unauthorized:
        return "($code) Unauthorized.$extraMessage";

      case HttpStatus.forbidden:
        return "($code) Forbidden.$extraMessage";

      case HttpStatus.notFound:
        return "($code) Not Found.$extraMessage";

      case HttpStatus.internalServerError:
        return "($code) Internal Server Error.$extraMessage";

      case AdeptUtils.Jsonable.jsonParserError:
        return "($code) Invalid response. JsonParser was unable to decode response.$extraMessage";

      default:
        return "HTTP error code: ($code).$extraMessage";
    }
  }
}
