import 'package:adeptarch_base/src/resource/resource.dart';
import 'package:adeptarch_base/src/dao/super_dao.dart';
import 'package:adeptarch_base/src/utils/adept_publish_subject.dart';
import 'dart:async';

abstract class SuperRepository<Dao extends SuperDao<Data>, ResourceData, Data,
    Param> implements ResourceDataAccess<ResourceData, Data, Param> {
  Dao dao;
  List<Param>? params;
  AdeptPublishSubject<Resource<ResourceData?>> publishSubject;
  AdeptPublishSubject<Resource<List<ResourceData>?>> publishSubjectList;

  SuperRepository(this.dao)
      : publishSubject = AdeptPublishSubject(),
        publishSubjectList = AdeptPublishSubject();

  AdeptPublishSubject<Resource<ResourceData?>> getObservableData() =>
      publishSubject;
  AdeptPublishSubject<Resource<List<ResourceData>?>> getObservableDataList() =>
      publishSubjectList;

  /// Loading
  Resource<ResourceData> postLoading(
      {ResourceData? resourceData,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    publishSubject ??= this.publishSubject;
    resourceData ??=
        publishSubject.value != null ? publishSubject.value!.data : null;
    final resource = Resource<ResourceData>.loading(resourceData);
    publishSubject.post(resource);
    return resource;
  }

  /// Success
  Resource<ResourceData> postSuccess(
      {ResourceData? resourceData,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    publishSubject ??= this.publishSubject;
    resourceData ??=
        publishSubject.value != null ? publishSubject.value!.data : null;
    final resource = Resource<ResourceData>.success(resourceData);
    publishSubject.post(resource);
    return resource;
  }

  /// Error
  Resource<ResourceData> postError(String? message,
      {ResourceData? resourceData,
      int? code,
      dynamic exception,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    publishSubject ??= this.publishSubject;
    resourceData ??=
        publishSubject.value != null ? publishSubject.value!.data : null;
    final resource = Resource<ResourceData>.error(resourceData, message,
        code: code, exception: exception);
    publishSubject.post(resource);
    return resource;
  }

  @override
  Data? loadSync({dynamic params}) => dao.load();

  @override
  Future<Data?> loadAsync({dynamic params}) => dao.loadAsync();

  @override
  AdeptPublishSubject<Resource<ResourceData?>> load(
      {List<Param>? params,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    publishSubject ??= this.publishSubject;
    if (this.params == params && Resource.hasData(publishSubject.value)) {
      postSuccess();
      return publishSubject;
    }
    return loadNew(params: params, publishSubject: publishSubject);
  }

  @override
  AdeptPublishSubject<Resource<ResourceData?>> loadNew(
      {List<Param>? params,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    publishSubject ??= this.publishSubject;
    return ResourceLoader<ResourceData?>(
        publishSubject: publishSubject,
        loadFromCache: () async {
          try {
            return dao.load() as ResourceData?;
          } catch (e) {
            //If there is an exception here means Data can't cast to ResourceData then this method should be Overridden
            print(e);
          }
          return null;
        }).asPublishSubject();
  }

  @override
  void save(
      {Data? data,
      List<Data>? dataList,
      bool? updateRef,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) async {
    if (data == null && dataList == null) return;
    publishSubject ??= this.publishSubject;
    postLoading();
    try {
      try {
        dao.save(data: data, dataList: dataList);
        postSuccess(
            resourceData: data as ResourceData?,
            publishSubject: publishSubject);
      } catch (e) {
        postError(e.toString(),
            resourceData: data as ResourceData?,
            publishSubject: publishSubject);
      }
    } catch (e) {
      //If there is an exception here means Data can't cast to ResourceData then this method should be Overridden
      print(e);
    }
  }

  @override
  void delete(
      {Data? data,
      List<Data>? dataList,
      bool? updateRef,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    if (data == null && dataList == null) return;
    publishSubject ??= this.publishSubject;
    postLoading(publishSubject: publishSubject);
    try {
      try {
        dao.delete(data: data, dataList: dataList);
        postSuccess(
            resourceData: data as ResourceData?,
            publishSubject: publishSubject);
      } catch (e) {
        postError(e.toString(),
            resourceData: data as ResourceData?,
            publishSubject: publishSubject);
      }
    } catch (e) {
      //If there is an exception here means Data can't cast to ResourceData then this method should be Overridden
      print(e);
    }
  }

  @override
  void clear(
      {bool? updateRef,
      AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) {
    dao.clear();
    (publishSubject ?? this.publishSubject)
        .post(Resource<ResourceData>.success(null));
  }

  @override
  void notifyChanges(
      {AdeptPublishSubject<Resource<ResourceData?>>? publishSubject}) async {
    publishSubject ??= this.publishSubject;
    publishSubject.post(publishSubject.value);
  }
}
